/*
 * Copyright 2010 Aalto University, ComNet
 * Released under GPLv3. See LICENSE.txt for details.
 */
package routing;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import core.Connection;
import core.DTNHost;
import core.Message;
import core.Settings;

/**
 * Implementation of Spray and wait router as depicted in
 * <I>Spray and Wait: An Efficient Routing Scheme for Intermittently
 * Connected Mobile Networks</I> by Thrasyvoulos Spyropoulus et al.
 *
 */
public class SprayAndWaitRouterGH extends ActiveRouter {
	/** identifier for the initial number of copies setting ({@value})*/
	public static final String NROF_COPIES = "nrofCopies";
	/** identifier for the binary-mode setting ({@value})*/
	public static final String BINARY_MODE = "binaryMode";
	/** SprayAndWait router's settings name space ({@value})*/
	public static final String SPRAYANDWAIT_NS = "SprayAndWaitRouterWithGH";
	/** Message property key */
	public static final String MSG_COUNT_PROPERTY = SPRAYANDWAIT_NS + "." +
		"copies";

	public static final String SPRAYANDWAIT_NS_GH = "SprayAndWaitRouterGH";
	/** Message property key */
	public static final String MSG_COUNT_PROPERTY_GH = SPRAYANDWAIT_NS_GH + "." +
		"copies";

	protected int initialNrofCopies;
	protected boolean isBinary;

	public SprayAndWaitRouterGH(Settings s) {
		super(s);
		Settings snwSettings = new Settings(SPRAYANDWAIT_NS);

		initialNrofCopies = snwSettings.getInt(NROF_COPIES);
		isBinary = snwSettings.getBoolean( BINARY_MODE);
	}

	/**
	 * Copy constructor.
	 * @param r The router prototype where setting values are copied from
	 */
	protected SprayAndWaitRouterGH(SprayAndWaitRouterGH r) {
		super(r);
		this.initialNrofCopies = r.initialNrofCopies;
		this.isBinary = r.isBinary;
	}

	@Override
	public int receiveMessage(Message m, DTNHost from) {
		return super.receiveMessage(m, from);
	}

	@Override
	public Message messageTransferred(String id, DTNHost from) {
		Message msg = super.messageTransferred(id, from);
		if (isBinary) {
			/* in binary S'n'W the receiving node gets floor(n/2) copies */
			try{				
				Integer nrofCopies = (Integer)msg.getProperty(MSG_COUNT_PROPERTY);

				assert nrofCopies != null : "Not a SnW message: " + msg;

				nrofCopies = (int)Math.floor(nrofCopies/2.0);

				msg.updateProperty(MSG_COUNT_PROPERTY, nrofCopies);

			} catch (Exception e) {
				Integer nrofCopies = (Integer)msg.getProperty(MSG_COUNT_PROPERTY_GH);

				assert nrofCopies != null : "Not a SnW message: " + msg;

				//nrofCopies = (int)Math.floor(nrofCopies/2.0);

				msg.updateProperty(MSG_COUNT_PROPERTY_GH, nrofCopies);
			}
		}
		return msg;		
	}

	@Override
	public boolean createNewMessage(Message msg) {
		makeRoomForNewMessage(msg.getSize());

		msg.setTtl(this.msgTtl);
		msg.addProperty(MSG_COUNT_PROPERTY, new Integer(initialNrofCopies));
		addToMessages(msg, true);
		return true;
	}

	@Override
	public void update() {
		super.update();
		if (!canStartTransfer() || isTransferring()) {
			return; // nothing to transfer or is currently transferring
		}

		Random rand = new Random();

		int randInt = rand.nextInt(100);

		if (randInt%2 == 1) {
			/* try messages that could be delivered to final recipient */
			if (exchangeDeliverableMessages() != null) {
				return;
			}

			/* create a list of SAWMessages that have copies left to distribute */
			@SuppressWarnings(value = "unchecked")
			List<Message> copiesLeft = sortByQueueMode(getMessagesWithCopiesLeft());

			if (copiesLeft.size() > 0) {
				/* try to send those messages */
				this.tryMessagesToConnections(copiesLeft, getConnections());
			}
		}
	}

	/**
	 * Creates and returns a list of messages this router is currently
	 * carrying and still has copies left to distribute (nrof copies > 1).
	 * @return A list of messages that have copies left
	 */
	protected List<Message> getMessagesWithCopiesLeft() {
		List<Message> list = new ArrayList<Message>();

		for (Message m : getMessageCollection()) {
			Integer nrofCopies = (Integer)m.getProperty(MSG_COUNT_PROPERTY);
			assert nrofCopies != null : "SnW message " + m + " didn't have " +
				"nrof copies property!";
			if (nrofCopies > 1) {
				list.add(m);
			}
		}

		return list;
	}

	/**
	 * Called just before a transfer is finalized (by
	 * {@link ActiveRouter#update()}).
	 * Reduces the number of copies we have left for a message.
	 * In binary Spray and Wait, sending host is left with floor(n/2) copies,
	 * but in standard mode, nrof copies left is reduced by one.
	 */
	@Override
	protected void transferDone(Connection con) {
		if (isBinary) {
			/* in binary S'n'W the sending node keeps ceil(n/2) copies */
			
			try{
				Integer nrofCopies;
				String msgId = con.getMessage().getId();
				/* get this router's copy of the message */
				Message msg = getMessage(msgId);

				if (msg == null) { // message has been dropped from the buffer after..
					return; // ..start of transfer -> no need to reduce amount of copies
				}

				
				nrofCopies = (Integer)msg.getProperty(MSG_COUNT_PROPERTY);

				/* reduce the amount of copies left */
				nrofCopies = (int)Math.ceil(nrofCopies/2.0);
				msg.updateProperty(MSG_COUNT_PROPERTY, nrofCopies);
			} catch (Exception e) {
				Integer nrofCopies;
				String msgId = con.getMessage().getId();
				/* get this router's copy of the message */
				Message msg = getMessage(msgId);

				if (msg == null) { // message has been dropped from the buffer after..
					return; // ..start of transfer -> no need to reduce amount of copies
				}

				
				nrofCopies = (Integer)msg.getProperty(MSG_COUNT_PROPERTY_GH);

				/* reduce the amount of copies left */
				//nrofCopies = (int)Math.ceil(nrofCopies/2.0);
				msg.updateProperty(MSG_COUNT_PROPERTY_GH, nrofCopies);
			}
		}
		
	}

	@Override
	public SprayAndWaitRouterGH replicate() {
		return new SprayAndWaitRouterGH(this);
	}
}
